#!/usr/bin/python
# file: cho2sgf.py
# Create SGF files containing the life and death problems in
# 'Cho's Encyclopedia of life and death' 
# This program takes the main problem collection coming with
# the encyclopedia (abc), translates it to SGF and writes the output
# into SGF files, one for each problem level (elementary, intermediate,
# advanced, others)

# Written by Ulrich Goertz (uligo@g0ertz.de), 2001

# The main part of this program is a translation (from C) of a very similar program
# written by Bill Shubert.

##   This program is free software; you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation; either version 2 of the License, or
##   (at your option) any later version.

##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.

##   You should have received a copy of the GNU General Public License
##   along with this program (gpl.txt); if not, write to the Free Software
##   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##   The GNU GPL is also currently available at
##   http://www.gnu.org/copyleft/gpl.html

# Of course, the Encyclopedia of Life and Death is copyrighted, and not freely distributable.
# It is available from Kiseido (http://www.kiseido.com/) or Het Paard
# (http://www.xs4all.nl/~paard/).

# -------------------------------------------------------------------------------------------

from Sgflib.sgflib import *
import sys

class Bunch: pass

class simpleString:
    def __init__(self): self.s = ''
    def append(self, a): self.s = self.s + a 
    def __str__(self): return self.s
    def __repr__(self): return self.s
    
def toInt(s):
    return ord(s[0]) + 256*ord(s[1]) + 256*256*ord(s[2]) + 256*256*256*ord(s[3])

def invert(color):
    if color == 'B' : return 'W'
    if color == 'W' : return 'B'



def addNode(nodes, pbm, prevNode, color, bestResult):
    numVar = 0   # number of variations starting from node prevNode
    for node in nodes:
        if node.prev==prevNode: numVar = numVar + 1

    if numVar:
        for i in range(len(nodes)):
            if nodes[i].prev == prevNode:
                if numVar>1: pbm.append('(')
                pbm.append(';' + color + '[' + chr(nodes[i].x + ord('a')) \
                                 + chr(nodes[i].y + ord('a')) + ']')
                if nodes[i].message % 16 :
                    pbm.append('C[' + getMessage(nodes[i].message, bestResult) + ']')
                addNode(nodes, pbm, i, invert(color), bestResult)
                if numVar>1: pbm.append(')\n')



def getMessage(message, bestResult):
    m = message / 16
    if m in [0, 8]:
        if message: return 'Correct.'
    elif m == 1 or m == 9:
        if bestResult == 1: return 'Correct; advantageous ko.'
        else: return 'Wrong; advantageous ko.'
    elif m == 2 or m == 10:
        if bestResult == 2: return 'Correct; disadvantageous ko.'
        else: return 'Wrong; disadvantageous ko.'
    elif m == 11:
        return 'Correct; ten thousand year ko.'
    elif m == 3:
        return 'Wrong.'
    else: return 'Unknown message ' + str(message%16) + ', ' + str(m) + ', ' + str(bestResult)


def readProblem(probStart):
    pbm = simpleString()
    pbm.append('(;GM[1]FF[4]\n')
    pbm.append('SZ[19]HA[0]KM[0]GN[Cho L&D (' + problemCollection +')]')

    # set up initial position

    pos = probStart

    pbm.append('AB')
    loc = -1    
    while ord(data[pos]):
        loc = loc + ord(data[pos])
        pbm.append('[' + chr(loc%19+ord('a')) + chr(loc/19+ord('a')) + ']')
        pos = pos + 1

    pos = pos + 1
    pbm.append('\nAW')
    loc = -1
    while ord(data[pos]) != 0:
        loc = loc + ord(data[pos])
        pbm.append('[' + chr(loc%19+ord('a')) + chr(loc/19+ord('a')) + ']')
        pos = pos + 1

    pbm.append('\n')
    pos = pos + 2

    x = ord(data[pos])

    if x == 42 or x == 74: pbm.append('C[Elementary]\n')
    elif x == 52 or x == 84: pbm.append('C[Intermediate]\n')
    elif x == 62 or x == 94 or x == 126: pbm.append('C[Advanced]\n')

    pos = pos + 1

    numNodes = ord(data[pos]) + 256*ord(data[pos+1])
    pos = pos + 2

    nodes = []
    for currentNode in range(numNodes):
        node = Bunch()
        node.prev = (ord(data[pos]) + 256*ord(data[pos+1])) % 2048
        node.x = ord(data[pos+1])/8
        node.y = ord(data[pos+2])
        node.message = ord(data[pos+3])
        # ignore data[pos+4]
        nodes.append(node)

        pos = pos + 5

    bestResult = 3
    for i in range(numNodes):
        if nodes[i].message:
            if (nodes[i].message / 16) % 8 < bestResult:
                bestResult = (nodes[i].message / 16) % 8
        
    addNode(nodes, pbm, 2047, 'B', bestResult)

    pbm.append(')\n')

    SGF = SGFParser(str(pbm)).parse()

    problem = SGF[0]

    cursor = problem.cursor()
    markWrongVar(cursor)
    
    if cursor.node.has_key('C'):
        print cursor.node['C'][0],
        print ''
        if cursor.node['C'][0] == 'Elementary': elemFile.write(str(problem)+'\n')
        elif cursor.node['C'][0] == 'Intermediate': intermedFile.write(str(problem)+'\n')
        elif cursor.node['C'][0] == 'Advanced': advFile.write(str(problem)+'\n')
        else: othersFile.write(str(problem)+'\n')
    else:
        othersFile.write(str(problem)+'\n')
        print ''




def markWrongVar(cursor):
    exWrong = 0
    exRight = 0

    allWrong = []
    for i in range(len(cursor.children)):
        c = cursor.next(i)
        if c.has_key('C') and c['C'][0][:5] == 'Wrong': allWrong.append(1)
        elif c.has_key('C') and c['C'][0][:7] == 'Correct' : allWrong.append(0)
        else: allWrong.append(markWrongVar(cursor))
        cursor.previous()
        
    if not 1 in allWrong: return 0
    if not 0 in allWrong: return 1

    for i in range(len(cursor.children)):
        cursor.next(i)
        if allWrong[i]: cursor.node.addProperty(Property('WV', None))
        cursor.previous()

    return 0


def printHelp():
    print 'This program translates the main problem collection from Cho\'s'
    print 'Encyclopedia of Life and Death to SGF format.'
    print 'It was written by Ulrich Goertz (uligo@g0ertz.de), 2001, and is'
    print 'based on a similar program written by Bill Shubert.'
    print 'This program comes WITHOUT ANY WARRANTY! Only use it with'
    print 'backups of the original files! This program is published under'
    print 'the GNU General Public License (see the file gpl.txt).'
    print ''
    print 'Usage:'
    print 'Copy this program into the directory with the files from'
    print 'the Encyclopedia (abc.dr and abc.dt are needed), and start it '
    print 'with \'python cho2sgf.py\'.'
    print ''
    print 'Four SGF files will be written (abc-el.sgf, abc-im.sgf, abc-adv.sgf,'
    print 'abc-oth.sgf; they contain the elementary, intermediate, advanced'
    print 'and not classified problems of the database, respectively.)'
    print 'They are ready to use with the uliGo program.'
    print ''
    print 'Of course, Cho\'s Encyclopedia of Life and Death is copyrighted and is'
    print 'not freely distributable. It is available from Kiseido'
    print '(http://www.kiseido.com) or from Het Paard (http://www.xs4all.nl/~paard/).'


if len(sys.argv) != 1:
    printHelp()
    sys.exit()

problemCollection = 'abc' # replace abc by s1 or s2 to try the other collections
                          # but some of the SGF files will be corrupt
                          # (I don't know why, but I didn't spend much time to find
                          # out, either.)

try:
    tableFile = open(problemCollection+'.dr', 'rb')
    table = tableFile.read()
    tableFile.close()

    dataFile = open(problemCollection+'.dt', 'rb')
    data = dataFile.read()
    dataFile.close()

    elemFile = open(problemCollection+'-el.sgf','w')
    elemFile.write('Elementary problems from Cho\'s Encyclopedia of Life and Death\n')
    intermedFile = open(problemCollection+'-im.sgf','w')
    intermedFile.write('Intermediate problems from Cho\'s Encyclopedia of Life and Death\n')
    advFile = open(problemCollection+'-adv.sgf','w')
    advFile.write('Advanced problems from Cho\'s Encyclopedia of Life and Death\n')
    othersFile = open(problemCollection+'-oth.sgf','w')
    othersFile.write('Miscellaneous problems from Cho\'s Encyclopedia of Life and Death\n')
except IOError:
    printHelp()
    sys.exit()

noOfProblems = len(table)/4

readProblem(0)

for i in range(0,noOfProblems-1):
    print 'Reading problem no ', str(i+1), ', ',
    probStart = toInt(table[4*i:4*i+4])
    readProblem(probStart)

elemFile.close()
intermedFile.close()
advFile.close()
othersFile.close()


    


